;;;; irc.scm - Build and maintain an irc connection -*- scheme -*-
;;;; Copyright (C) 2001  Neal H Walfield  <neal@cs.uml.edu>
;;;;
;;;; This file is part of Deride
;;;;
;;;; Deride is free software; you can redistribute it and/or modify
;;;; it under the terms of the GNU General Public License as published by
;;;; the Free Software Foundation; either version 2, or (at your option)
;;;; any later version.
;;;; 
;;;; Deride is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; GNU General Public License for more details.
;;;; 
;;;; You should have received a copy of the GNU General Public License
;;;; along with the Deride; see the file COPYING.  If not, write to
;;;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

(define-module (comm irc))

(use-modules (ice-9 slib))
(use-modules (ice-9 regex))

(use-modules (comm irc rfc-2812))
(use-modules (comm irc rfc-2812-extensions))

(use-modules (forum users))

(use-modules (comm hooks))

;; The end of line marker for irc servers.
(define-public irc-eol "\n\r")

;; The user's nick.
(define-public irc-nick "")

;; Connect to SERVER on PORT.  Return a SOCKET.
(define-public (irc-connect server port)
  (let*
      ((s (socket AF_INET SOCK_STREAM 0))
       (host (catch
	      #t
	      (lambda () (hostent:addr-list (gethostbyname server)))
	      (lambda args
		(run-hook comm-global-status-message-hook
			  "error" (strerror (car (list-ref args 4))))
		#f))))
    (and host
	 (catch
	  #t
	  (lambda () (connect s AF_INET (car host) port))
	  (lambda args
	    (run-hook comm-global-status-message-hook
		      "error" (strerror (car (list-ref args 4))))
	    #f))
	 s)))

;; Send DATA to SOCKET and the STDOUT
(define-public (irc-send-raw data socket)
;  (display (string-append data "\n"))
  (write-line (string-append data irc-eol) socket))

;; Login to the irc server on SOCKET using NICK and REALNAME
(define-public (irc-send-login socket nick realname)
  (set! irc-nick nick)
  (irc-send-raw (string-append "USER " nick " 0 * :" realname)
		       socket)
  (irc-send-raw (string-append "NICK " nick) socket))

;; Outstanding number of pings to the server
(define irc-outstanding-server-pings 0)

;; Ping the server.
(define (irc-ping-server socket)
  (irc-send-raw (string-append "PING " irc-server-name) socket)
  (set! irc-outstanding-server-pings (1+ irc-outstanding-server-pings)))

;; The name of the irc server we are directly connected to (as it sees
;; fit to call itself).
(define irc-server-name "unknown")

;; Parse a packet and act on it!
(define (parse-packet data socket)
  ;; Parse an irc message RFC1459, section 2.3.1
  (define irc-message-regex
    (make-regexp
     (string-append
      "^(:(([^ !]+)!?([^ ]*)))? *"      ; The prefix: i.e. `:nick!user@host'
      "([a-zA-Z]+|[0-9]{3}) +"          ; The command
      "(([^:]*):?)?")))                 ; The middle parameters
					; Trailing is in the suffix

  (let ((return-code #t)
	(rx (regexp-exec irc-message-regex data)))
    (if (not rx)
	(run-hook comm-global-status-message-hook
		  "error" (string-append "Unparsable message from server: `"
					 data "'"))
	(let* ((prefix (or (match:substring rx 2) ""))
	       (user (or (match:substring rx 3) ""))
	       (users-host (or (match:substring rx 4) ""))
	       (command (or (match:substring rx 5) ""))
	       (middle (or (match:substring rx 7) ""))
	       (trailing (or (match:suffix rx) ""))

	       ;; Break the first word off of a string (spaces are
	       ;; considered seperators); Returns a the word and
	       ;; modifies the string such that it contains the the
	       ;; rest of the string (with any leading spaces stripped
	       ;; off).
	       (get-word
		(lambda ()
		  (let* ((start (string-index middle #\space))
			 (word (if start
				   (substring middle 0 start)
				   middle))
			 (strip-leading-space
			  (lambda (s)
			    (if (or (= (string-length s) 0)
				    (not (eqv? (string-ref s 0) #\space)))
				s
				(strip-leading-space (substring s 1))))))
		    (set! middle (if start
				     (strip-leading-space
				      (substring middle (1+ start))) ""))
		    word)))

	       (first-middle (get-word))
	       (second-middle (get-word))
	       (third-middle (get-word))
	       (fourth-middle (get-word)))

	  (cond

	   ;; Message.
	   ;; By far the most common message.  Ergo, it is at the top.
	   ((or
	     (string-ci=? command "PRIVMSG")
	     (string-ci=? command "NOTICE"))
	    (let* ((say (and
			 (>= (string-length trailing) 7)
			 (= (char->integer (string-ref trailing 0)) 1)
			 (string=? "ACTION " (substring trailing 1 8))))
		   (msg (if say
			    (substring trailing 8)
			    trailing)))
	      (if (string-ci=? first-middle irc-nick)
		  (run-hook comm-received-personal-message-hook
			    user say msg)
		  (run-hook comm-received-forum-message-hook
			    user first-middle say msg))))

	   ;; A user joined a channel
	   ((string-ci=? command "JOIN")
	    (forum-user-joined trailing user ""))

	   ;; A user left a channel
	   ((string-ci=? command "PART")
	    (forum-user-parted first-middle user trailing))

	   ;; A user has quit.  Hm, it does not give us the channels that he
	   ;; was on.  Looks like we need a reverse map.
	   ((string-ci=? command "QUIT")
	    (forum-user-quit user trailing))

	   ;; Same for NICK.
	   ((string-ci=? command "NICK")
	    (forum-user-schizophrenia user trailing ""))

	   ;; Just got a list of people in the channel
	   ((string-ci=? command RPL_NAMREPLY)
	    (letrec ((split (lambda (string)
			    (if (> (string-length string) 0)
				(case (string-ref string 0)
				  ((#\space #\@ #\+)
				   (split (substring string 1)))
				  (else
				   (let ((end (string-index string #\space)))
				     (forum-user-joined
				      third-middle
				      (substring string 0 end)
				      "")
				     (split (substring string (1+ end))))))))))
	      (split trailing)))

	   ;; End of the list of nicks.
#!
	   ((string=? command RPL_ENDOFNAMES)
	    (client-display (string-append
			     "Users on " second-middle ": "
			     (list-of-strings->string
			      (forum-user-list second-middle) ", " " and ")
			     "\n")))
!#
	   ;; PING!
	   ((string-ci=? command "PING")
	    (irc-send-raw (string-append
			   "PONG " (substring data 6))
			  socket))

	   ;; PONG!
	   ((string-ci=? command "PONG")
	    (if (string-ci=? prefix irc-server-name)
		;; The server responded to a ping, mark it.  This
		;; makes the pretty reasonable assumption that the
		;; only way we will ever receive a PONG from our
		;; server is in response to a successful call to
		;; irc-ping-server (which incremented
		;; irc-outstanding-server-pings).
		(set! irc-outstanding-server-pings
		      (1- irc-outstanding-server-pings))))

	   ;; Welcome to the irc server, blah, blah, blah.  Save the
	   ;; server's name so we can ping it later; this is the easiest
	   ;; spot to get it.
	   ((or
	     (string=? command RPL_WELCOME)
	     (string=? command RPL_YOURHOST)
	     (string=? command RPL_CREATED)
	     (string=? command RPL_MYINFO))
	    (set! irc-server-name prefix)
	    (run-hook comm-global-status-message-hook prefix trailing))

	   ;; Server info; just echo it.
	   ((or
	     (string=? command RPL_LUSERCLIENT)
	     (string=? command RPL_LUSEROP)
	     (string=? command RPL_LUSERUNKNOWN)
	     (string=? command RPL_LUSERCHANNELS)
	     (string=? command RPL_LUSERME)
	     (string=? command RPL_LLOCALUSERS)
	     (string=? command RPL_LGLOBALUSERS)
	     (string=? command RPL_LCONNECTIONCOUNT))
	    (run-hook comm-global-status-message-hook second-middle trailing))

	   ;; Try to connect to a different server.
	   ((string=? command RPL_BOUNCE)
	    (run-hook comm-global-status-message-hook
		      prefix (string-append "Server requests a reconnect to:"
					    trailing))
	    (set! return-code #f))

	   ;; Initial topic.
	   ((string=? command RPL_TOPIC)
	    (run-hook comm-forum-status-message-hook
		      second-middle
		      (string-append "Topic is: " trailing)))

	   ((string=? command RPL_TOPICWHOTIME)
	    (run-hook comm-forum-status-message-hook second-middle
		      (string-append "Topic set by "
				     third-middle " on "
				     (strftime
				      "%c"
				      (localtime
				       (or (string->number fourth-middle) 0)))
				     "\n")))

	   ;; Topic change.
	   ((string=? command "TOPIC")
	    (run-hook comm-forum-status-message-hook first-middle
		      (string-append user " changed the topic to: " trailing)))

	   ;; MOTD
	   ((or (string=? command RPL_MOTDSTART)
		(string=? command RPL_MOTD)
		(string=? command RPL_ENDOFMOTD))
	    (run-hook comm-global-status-message-hook
		      "motd" trailing))

	   ((run-hook comm-global-status-message-hook
		      "unknown command" data)))))
    return-code))

;; Ping our server after having not received a message for this many
;; seconds.
(define max-silence 200)

;; Enter the server loop.  Read a line from SOCKET and pass it to
;; parse-packet.  If parse-packet returns false or EOF of file on SOCKET,
;; is reached, return.  Otherwise, loop.
(define-public (irc-loop socket)
  (if (eqv? (car (select (list (port->fdes socket)) '() '() max-silence))
	    '())
      (if (> irc-outstanding-server-pings 0)
	  (begin (run-hook comm-global-status-message-hook
			   "error" "Ping timeout.")
		 (if (not (port-closed? socket))
		     close-port socket))
	  (and (irc-ping-server socket)
	       (irc-loop socket)))
      (let ((data 
	     (catch
	      #t
	      (lambda () (read-line socket))
	      (lambda args
		(run-hook comm-global-status-message-hook
			  "error" (strerror (car (list-ref args 4))))
		#f))))
	(if data
	    (or (eof-object? data)
		(not (parse-packet (substring data 0 (1- (string-length data)))
				   socket))
		(irc-loop socket))
	    (run-hook comm-global-status-message-hook
		      "error" "Socket read error.")))))
